/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.nd4j.linalg.api.ops.impl.scalar;

import org.nd4j.autodiff.samediff.SDVariable;
import org.nd4j.autodiff.samediff.SameDiff;
import org.nd4j.imports.NoOpNameFoundException;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.api.ops.BaseScalarOp;

import java.util.List;

public class ReplaceNans extends BaseScalarOp {

    private double set;

    public ReplaceNans(SameDiff sameDiff, SDVariable i_v, boolean inPlace, double set) {
        super(sameDiff, i_v, set, inPlace);
        this.set = set;
    }

    public ReplaceNans(SameDiff sameDiff, SDVariable i_v, Object[] extraArgs, double set) {
        super(sameDiff, i_v, set, extraArgs);
        this.set = set;
    }

    public ReplaceNans() {

    }

    public ReplaceNans(INDArray x, double set) {
        super(x, null, x, set);
        this.set = set;
        this.extraArgs = new Object[] {set};
    }

    public ReplaceNans(INDArray x, INDArray z, double set) {
        super(x, z, set);
        this.set = set;
        this.extraArgs = new Object[] {set};
    }

    @Override
    public int opNum() {
        return 37;
    }

    @Override
    public String opName() {
        return "replace_nans";
    }

    @Override
    public String onnxName() {
        throw new NoOpNameFoundException("No onnx op opName found for " +  opName());
    }

    @Override
    public String tensorflowName() {
        throw new NoOpNameFoundException("No tensorflow op opName found for " +  opName());
    }

    @Override
    public List<SDVariable> doDiff(List<SDVariable> f1) {
        return null;
    }
}

